@@ -1,51 +1,42 @@
 /* pinentry.c - The PIN entry support library
-   Copyright (C) 2002, 2003, 2007, 2008, 2010, 2015, 2016 g10 Code GmbH
-
-   This file is part of PINENTRY.
-
-   PINENTRY is free software; you can redistribute it and/or modify it
-   under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
-
-   PINENTRY is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, see <https://www.gnu.org/licenses/>.
+ * Copyright (C) 2002, 2003, 2007, 2008, 2010, 2015, 2016, 2021 g10 Code GmbH
+ *
+ * This file is part of PINENTRY.
+ *
+ * PINENTRY is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * PINENTRY is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, see <https://www.gnu.org/licenses/>.
+ * SPDX-License-Identifier: GPL-2.0+
  */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 
-#ifndef HAVE_W32CE_SYSTEM
-# include <errno.h>
-#endif
+#include <errno.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #include <unistd.h>
 #include <assert.h>
-#ifndef HAVE_W32CE_SYSTEM
-# include <locale.h>
-#endif
-#ifdef HAVE_LANGINFO_H
-#include <langinfo.h>
+#ifndef HAVE_W32_SYSTEM
+# include <sys/utsname.h>
 #endif
+#include <locale.h>
 #include <limits.h>
-#ifdef HAVE_W32CE_SYSTEM
-# include <windows.h>
-#endif
-
-#if defined FALLBACK_CURSES || defined PINENTRY_CURSES || defined PINENTRY_GTK
-#include <iconv.h>
-#endif
 
 #include <assuan.h>
 
-#include "memory.h"
 #include "secmem-util.h"
 #include "argparse.h"
 #include "pinentry.h"
@@ -58,15 +49,23 @@
 # include "pinentry-curses.h"
 #endif
 
-#ifdef HAVE_W32CE_SYSTEM
-#define getpid() GetCurrentProcessId ()
-#endif
 
 /* Keep the name of our program here. */
 static char this_pgmname[50];
 
 struct pinentry pinentry;
 
+
+static const char *flavor_flag;
+
+/* Because gtk_init removes the --display arg from the command lines
+ * and our command line parser is called after gtk_init (so that it
+ * does not see gtk specific options) we don't have a way to get hold
+ * of the --display option.  Our solution is to remember --display in
+ * the call to pinentry_have_display and set it then in our
+ * parser.  */
+static char *remember_display;
+
 static void
 pinentry_reset (int use_defaults)
 {
@@ -74,7 +73,8 @@ pinentry_reset (int use_defaults)
      Don't reset them.  */
   int grab = pinentry.grab;
   char *ttyname = pinentry.ttyname;
-  char *ttytype = pinentry.ttytype;
+  char *ttytype = pinentry.ttytype_l;
+  char *ttyalert = pinentry.ttyalert;
   char *lc_ctype = pinentry.lc_ctype;
   char *lc_messages = pinentry.lc_messages;
   int allow_external_password_cache = pinentry.allow_external_password_cache;
@@ -85,7 +85,15 @@ pinentry_reset (int use_defaults)
   char *default_cf_visi = pinentry.default_cf_visi;
   char *default_tt_visi = pinentry.default_tt_visi;
   char *default_tt_hide = pinentry.default_tt_hide;
+  char *default_capshint = pinentry.default_capshint;
   char *touch_file = pinentry.touch_file;
+  unsigned long owner_pid = pinentry.owner_pid;
+  int owner_uid = pinentry.owner_uid;
+  char *owner_host = pinentry.owner_host;
+  int constraints_enforce = pinentry.constraints_enforce;
+  char *constraints_hint_short = pinentry.constraints_hint_short;
+  char *constraints_hint_long = pinentry.constraints_hint_long;
+  char *constraints_error_title = pinentry.constraints_error_title;
 
   /* These options are set from the command line.  Don't reset
      them.  */
@@ -98,6 +106,10 @@ pinentry_reset (int use_defaults)
   pinentry_color_t color_bg = pinentry.color_bg;
   pinentry_color_t color_so = pinentry.color_so;
   int color_so_bright = pinentry.color_so_bright;
+  pinentry_color_t color_ok = pinentry.color_ok;
+  int color_ok_bright = pinentry.color_ok_bright;
+  pinentry_color_t color_qualitybar = pinentry.color_qualitybar;
+  int color_qualitybar_bright = pinentry.color_qualitybar_bright;
 
   int timeout = pinentry.timeout;
 
@@ -108,7 +120,8 @@ pinentry_reset (int use_defaults)
   if (use_defaults)
     {
       free (pinentry.ttyname);
-      free (pinentry.ttytype);
+      free (pinentry.ttytype_l);
+      free (pinentry.ttyalert);
       free (pinentry.lc_ctype);
       free (pinentry.lc_messages);
       free (pinentry.default_ok);
@@ -118,8 +131,13 @@ pinentry_reset (int use_defaults)
       free (pinentry.default_cf_visi);
       free (pinentry.default_tt_visi);
       free (pinentry.default_tt_hide);
+      free (pinentry.default_capshint);
       free (pinentry.touch_file);
+      free (pinentry.owner_host);
       free (pinentry.display);
+      free (pinentry.constraints_hint_short);
+      free (pinentry.constraints_hint_long);
+      free (pinentry.constraints_error_title);
     }
 
   free (pinentry.title);
@@ -134,6 +152,7 @@ pinentry_reset (int use_defaults)
   free (pinentry.repeat_error_string);
   free (pinentry.quality_bar);
   free (pinentry.quality_bar_tt);
+  free (pinentry.formatted_passphrase_hint);
   free (pinentry.keyinfo);
   free (pinentry.specific_err_info);
 
@@ -158,13 +177,19 @@ pinentry_reset (int use_defaults)
       pinentry.color_bg = PINENTRY_COLOR_DEFAULT;
       pinentry.color_so = PINENTRY_COLOR_DEFAULT;
       pinentry.color_so_bright = 0;
+      pinentry.color_ok = PINENTRY_COLOR_DEFAULT;
+      pinentry.color_ok_bright = 0;
+      pinentry.color_qualitybar = PINENTRY_COLOR_DEFAULT;
+      pinentry.color_qualitybar_bright = 0;
+
+      pinentry.owner_uid = -1;
     }
-  else
-    /* Restore the options.  */
+  else /* Restore the options.  */
     {
       pinentry.grab = grab;
       pinentry.ttyname = ttyname;
-      pinentry.ttytype = ttytype;
+      pinentry.ttytype_l = ttytype;
+      pinentry.ttyalert = ttyalert;
       pinentry.lc_ctype = lc_ctype;
       pinentry.lc_messages = lc_messages;
       pinentry.allow_external_password_cache = allow_external_password_cache;
@@ -175,7 +200,15 @@ pinentry_reset (int use_defaults)
       pinentry.default_cf_visi = default_cf_visi;
       pinentry.default_tt_visi = default_tt_visi;
       pinentry.default_tt_hide = default_tt_hide;
+      pinentry.default_capshint = default_capshint;
       pinentry.touch_file = touch_file;
+      pinentry.owner_pid = owner_pid;
+      pinentry.owner_uid = owner_uid;
+      pinentry.owner_host = owner_host;
+      pinentry.constraints_enforce = constraints_enforce;
+      pinentry.constraints_hint_short = constraints_hint_short;
+      pinentry.constraints_hint_long = constraints_hint_long;
+      pinentry.constraints_error_title = constraints_error_title;
 
       pinentry.debug = debug;
       pinentry.display = display;
@@ -186,6 +219,10 @@ pinentry_reset (int use_defaults)
       pinentry.color_bg = color_bg;
       pinentry.color_so = color_so;
       pinentry.color_so_bright = color_so_bright;
+      pinentry.color_ok = color_ok;
+      pinentry.color_ok_bright = color_ok_bright;
+      pinentry.color_qualitybar = color_qualitybar;
+      pinentry.color_qualitybar_bright = color_qualitybar_bright;
 
       pinentry.timeout = timeout;
     }
@@ -204,149 +241,6 @@ pinentry_assuan_reset_handler (assuan_context_t ctx, char *line)
 
 
 
-static int lc_ctype_unknown_warning = 0;
-
-#if defined FALLBACK_CURSES || defined PINENTRY_CURSES || defined PINENTRY_GTK
-char *
-pinentry_utf8_to_local (const char *lc_ctype, const char *text)
-{
-  iconv_t cd;
-  const char *input = text;
-  size_t input_len = strlen (text) + 1;
-  char *output;
-  size_t output_len;
-  char *output_buf;
-  size_t processed;
-  char *old_ctype;
-  char *target_encoding;
-
-  /* If no locale setting could be determined, simply copy the
-     string.  */
-  if (!lc_ctype)
-    {
-      if (! lc_ctype_unknown_warning)
-	{
-	  fprintf (stderr, "%s: no LC_CTYPE known - assuming UTF-8\n",
-		   this_pgmname);
-	  lc_ctype_unknown_warning = 1;
-	}
-      return strdup (text);
-    }
-
-  old_ctype = strdup (setlocale (LC_CTYPE, NULL));
-  if (!old_ctype)
-    return NULL;
-  setlocale (LC_CTYPE, lc_ctype);
-  target_encoding = nl_langinfo (CODESET);
-  if (!target_encoding)
-    target_encoding = "?";
-  setlocale (LC_CTYPE, old_ctype);
-  free (old_ctype);
-
-  /* This is overkill, but simplifies the iconv invocation greatly.  */
-  output_len = input_len * MB_LEN_MAX;
-  output_buf = output = malloc (output_len);
-  if (!output)
-    return NULL;
-
-  cd = iconv_open (target_encoding, "UTF-8");
-  if (cd == (iconv_t) -1)
-    {
-      fprintf (stderr, "%s: can't convert from UTF-8 to %s: %s\n",
-               this_pgmname, target_encoding, strerror (errno));
-      free (output_buf);
-      return NULL;
-    }
-  processed = iconv (cd, (ICONV_CONST char **)&input, &input_len,
-                     &output, &output_len);
-  iconv_close (cd);
-  if (processed == (size_t) -1 || input_len)
-    {
-      fprintf (stderr, "%s: error converting from UTF-8 to %s: %s\n",
-               this_pgmname, target_encoding, strerror (errno));
-      free (output_buf);
-      return NULL;
-    }
-  return output_buf;
-}
-
-/* Convert TEXT which is encoded according to LC_CTYPE to UTF-8.  With
-   SECURE set to true, use secure memory for the returned buffer.
-   Return NULL on error. */
-char *
-pinentry_local_to_utf8 (char *lc_ctype, char *text, int secure)
-{
-  char *old_ctype;
-  char *source_encoding;
-  iconv_t cd;
-  const char *input = text;
-  size_t input_len = strlen (text) + 1;
-  char *output;
-  size_t output_len;
-  char *output_buf;
-  size_t processed;
-
-  /* If no locale setting could be determined, simply copy the
-     string.  */
-  if (!lc_ctype)
-    {
-      if (! lc_ctype_unknown_warning)
-	{
-	  fprintf (stderr, "%s: no LC_CTYPE known - assuming UTF-8\n",
-		   this_pgmname);
-	  lc_ctype_unknown_warning = 1;
-	}
-      output_buf = secure? secmem_malloc (input_len) : malloc (input_len);
-      if (output_buf)
-        strcpy (output_buf, input);
-      return output_buf;
-    }
-
-  old_ctype = strdup (setlocale (LC_CTYPE, NULL));
-  if (!old_ctype)
-    return NULL;
-  setlocale (LC_CTYPE, lc_ctype);
-  source_encoding = nl_langinfo (CODESET);
-  setlocale (LC_CTYPE, old_ctype);
-  free (old_ctype);
-
-  /* This is overkill, but simplifies the iconv invocation greatly.  */
-  output_len = input_len * MB_LEN_MAX;
-  output_buf = output = secure? secmem_malloc (output_len):malloc (output_len);
-  if (!output)
-    return NULL;
-
-  cd = iconv_open ("UTF-8", source_encoding);
-  if (cd == (iconv_t) -1)
-    {
-      fprintf (stderr, "%s: can't convert from %s to UTF-8: %s\n",
-               this_pgmname, source_encoding? source_encoding : "?",
-               strerror (errno));
-      if (secure)
-        secmem_free (output_buf);
-      else
-        free (output_buf);
-      return NULL;
-    }
-  processed = iconv (cd, (ICONV_CONST char **)&input, &input_len,
-                     &output, &output_len);
-  iconv_close (cd);
-  if (processed == (size_t) -1 || input_len)
-    {
-      fprintf (stderr, "%s: error converting from %s to UTF-8: %s\n",
-               this_pgmname, source_encoding? source_encoding : "?",
-               strerror (errno));
-      if (secure)
-        secmem_free (output_buf);
-      else
-        free (output_buf);
-      return NULL;
-    }
-  return output_buf;
-}
-#endif
-
-
 /* Copy TEXT or TEXTLEN to BUFFER and escape as required.  Return a
    pointer to the end of the new buffer.  Note that BUFFER must be
    large enough to keep the entire text; allocataing it 3 times of
@@ -374,6 +268,177 @@ copy_and_escape (char *buffer, const void *text, size_t textlen)
 }
 
 
+/* Perform percent unescaping in STRING and return the new valid length
+   of the string.  A terminating Nul character is inserted at the end of
+   the unescaped string.
+ */
+static size_t
+do_unescape_inplace (char *s)
+{
+  unsigned char *p, *p0;
+
+  p = p0 = s;
+  while (*s)
+    {
+      if (*s == '%' && s[1] && s[2])
+        {
+          s++;
+          *p++ = xtoi_2 (s);
+          s += 2;
+        }
+      else
+        *p++ = *s++;
+    }
+  *p = 0;
+
+  return (p - p0);
+}
+
+
+/* Return a malloced copy of the commandline for PID.  If this is not
+ * possible NULL is returned.  */
+#ifndef HAVE_W32_SYSTEM
+static char *
+get_cmdline (unsigned long pid)
+{
+  char buffer[200];
+  FILE *fp;
+  size_t i, n;
+
+  snprintf (buffer, sizeof buffer, "/proc/%lu/cmdline", pid);
+
+  fp = fopen (buffer, "rb");
+  if (!fp)
+    return NULL;
+  n = fread (buffer, 1, sizeof buffer - 1, fp);
+  if (n < sizeof buffer -1 && ferror (fp))
+    {
+      /* Some error occurred.  */
+      fclose (fp);
+      return NULL;
+    }
+  fclose (fp);
+  if (n == 0)
+    return NULL;
+  /* Arguments are delimited by Nuls.  We should do proper quoting but
+   * that can be a bit complicated, thus we simply replace the Nuls by
+   * spaces.  */
+  for (i=0; i < n; i++)
+    if (!buffer[i] && i < n-1)
+      buffer[i] = ' ';
+  buffer[i] = 0; /* Make sure the last byte is the string terminator.  */
+
+  return strdup (buffer);
+}
+#endif /*!HAVE_W32_SYSTEM*/
+
+
+/* Atomically ask the kernel for information about process PID.
+ * Return a malloc'ed copy of the process name as long as the process
+ * uid matches UID.  If it cannot determine that the process has uid
+ * UID, it returns NULL.
+ *
+ * This is not as informative as get_cmdline, but it verifies that the
+ * process does belong to the user in question.
+ */
+#ifndef HAVE_W32_SYSTEM
+static char *
+get_pid_name_for_uid (unsigned long pid, int uid)
+{
+  char buffer[400];
+  FILE *fp;
+  size_t end, n;
+  char *uidstr;
+
+  snprintf (buffer, sizeof buffer, "/proc/%lu/status", pid);
+
+  fp = fopen (buffer, "rb");
+  if (!fp)
+    return NULL;
+  n = fread (buffer, 1, sizeof buffer - 1, fp);
+  if (n < sizeof buffer -1 && ferror (fp))
+    {
+      /* Some error occurred.  */
+      fclose (fp);
+      return NULL;
+    }
+  fclose (fp);
+  if (n == 0)
+    return NULL;
+  buffer[n] = 0;
+  /* Fixme: Is it specified that "Name" is always the first line?  For
+   * robustness I would prefer to have a real parser here. -wk  */
+  if (strncmp (buffer, "Name:\t", 6))
+    return NULL;
+  end = strcspn (buffer + 6, "\n") + 6;
+  buffer[end] = 0;
+
+  /* check that uid matches what we expect */
+  uidstr = strstr (buffer + end + 1, "\nUid:\t");
+  if (!uidstr)
+    return NULL;
+  if (atoi (uidstr + 6) != uid)
+    return NULL;
+
+  return strdup (buffer + 6);
+}
+#endif /*!HAVE_W32_SYSTEM*/
+
+
+const char *
+pinentry_get_pgmname (void)
+{
+  return this_pgmname;
+}
+
+
+/* Return a malloced string with the title.  The caller mus free the
+ * string.  If no title is available or the title string has an error
+ * NULL is returned.  */
+char *
+pinentry_get_title (pinentry_t pe)
+{
+  char *title;
+
+  if (pe->title)
+    title = strdup (pe->title);
+#ifndef HAVE_W32_SYSTEM
+  else if (pe->owner_pid)
+    {
+      char buf[200];
+      struct utsname utsbuf;
+      char *pidname = NULL;
+      char *cmdline = NULL;
+
+      if (pe->owner_host &&
+          !uname (&utsbuf) &&
+          !strcmp (utsbuf.nodename, pe->owner_host))
+        {
+          pidname = get_pid_name_for_uid (pe->owner_pid, pe->owner_uid);
+          if (pidname)
+            cmdline = get_cmdline (pe->owner_pid);
+        }
+
+      if (pe->owner_host && (cmdline || pidname))
+        snprintf (buf, sizeof buf, "[%lu]@%s (%s)",
+                  pe->owner_pid, pe->owner_host, cmdline ? cmdline : pidname);
+      else if (pe->owner_host)
+        snprintf (buf, sizeof buf, "[%lu]@%s",
+                  pe->owner_pid, pe->owner_host);
+      else
+        snprintf (buf, sizeof buf, "[%lu] <unknown host>",
+                  pe->owner_pid);
+      free (pidname);
+      free (cmdline);
+      title = strdup (buf);
+    }
+#endif /*!HAVE_W32_SYSTEM*/
+  else
+    title = strdup (this_pgmname);
+
+  return title;
+}
+
 
 /* Run a quality inquiry for PASSPHRASE of LENGTH.  (We need LENGTH
    because not all backends might be able to return a proper
@@ -450,6 +515,122 @@ pinentry_inq_quality (pinentry_t pin, const char *passphrase, size_t length)
 }
 
 
+/* Run a checkpin inquiry */
+char *
+pinentry_inq_checkpin (pinentry_t pin, const char *passphrase, size_t length)
+{
+  assuan_context_t ctx = pin->ctx_assuan;
+  const char prefix[] = "INQUIRE CHECKPIN ";
+  char *command;
+  char *line;
+  size_t linelen;
+  int gotvalue = 0;
+  char *value = NULL;
+  int rc;
+
+  if (!ctx)
+    return 0; /* Can't run the callback.  */
+
+  if (length > 300)
+    length = 300;  /* Limit so that it definitely fits into an Assuan
+                      line.  */
+
+  command = secmem_malloc (strlen (prefix) + 3*length + 1);
+  if (!command)
+    return 0;
+  strcpy (command, prefix);
+  copy_and_escape (command + strlen(command), passphrase, length);
+  rc = assuan_write_line (ctx, command);
+  secmem_free (command);
+  if (rc)
+    {
+      fprintf (stderr, "ASSUAN WRITE LINE failed: rc=%d\n", rc);
+      return 0;
+    }
+
+  for (;;)
+    {
+      do
+        {
+          rc = assuan_read_line (ctx, &line, &linelen);
+          if (rc)
+            {
+              fprintf (stderr, "ASSUAN READ LINE failed: rc=%d\n", rc);
+              return 0;
+            }
+        }
+      while (*line == '#' || !linelen);
+      if (line[0] == 'E' && line[1] == 'N' && line[2] == 'D'
+          && (!line[3] || line[3] == ' '))
+        break; /* END command received*/
+      if (line[0] == 'C' && line[1] == 'A' && line[2] == 'N'
+          && (!line[3] || line[3] == ' '))
+        break; /* CAN command received*/
+      if (line[0] == 'E' && line[1] == 'R' && line[2] == 'R'
+          && (!line[3] || line[3] == ' '))
+        break; /* ERR command received*/
+      if (line[0] != 'D' || line[1] != ' ' || linelen < 3 || gotvalue)
+        continue;
+      gotvalue = 1;
+      value = strdup (line + 2);
+    }
+
+  return value;
+}
+
+
+/* Run a genpin inquiry */
+char *
+pinentry_inq_genpin (pinentry_t pin)
+{
+  assuan_context_t ctx = pin->ctx_assuan;
+  const char prefix[] = "INQUIRE GENPIN";
+  char *line;
+  size_t linelen;
+  int gotvalue = 0;
+  char *value = NULL;
+  int rc;
+
+  if (!ctx)
+    return 0; /* Can't run the callback.  */
+
+  rc = assuan_write_line (ctx, prefix);
+  if (rc)
+    {
+      fprintf (stderr, "ASSUAN WRITE LINE failed: rc=%d\n", rc);
+      return 0;
+    }
+
+  for (;;)
+    {
+      do
+        {
+          rc = assuan_read_line (ctx, &line, &linelen);
+          if (rc)
+            {
+              fprintf (stderr, "ASSUAN READ LINE failed: rc=%d\n", rc);
+              free (value);
+              return 0;
+            }
+        }
+      while (*line == '#' || !linelen);
+      if (line[0] == 'E' && line[1] == 'N' && line[2] == 'D'
+          && (!line[3] || line[3] == ' '))
+        break; /* END command received*/
+      if (line[0] == 'C' && line[1] == 'A' && line[2] == 'N'
+          && (!line[3] || line[3] == ' '))
+        break; /* CAN command received*/
+      if (line[0] == 'E' && line[1] == 'R' && line[2] == 'R'
+          && (!line[3] || line[3] == ' '))
+        break; /* ERR command received*/
+      if (line[0] != 'D' || line[1] != ' ' || linelen < 3 || gotvalue)
+        continue;
+      gotvalue = 1;
+      value = strdup (line + 2);
+    }
+
+  return value;
+}
 
 /* Try to make room for at least LEN bytes in the pinentry.  Returns
    new buffer on success and 0 on failure or when the old buffer is
@@ -566,17 +747,48 @@ pinentry_init (const char *pgmname)
 int
 pinentry_have_display (int argc, char **argv)
 {
-#ifndef HAVE_W32CE_SYSTEM
-  const char *s;
+  int found = 0;
 
-  s = getenv ("DISPLAY");
-  if (s && *s)
-    return 1;
-#endif
   for (; argc; argc--, argv++)
-    if (!strcmp (*argv, "--display") || !strncmp (*argv, "--display=", 10))
-      return 1;
-  return 0;
+    {
+      if (!strcmp (*argv, "--display"))
+        {
+          if (argv[1] && !remember_display)
+            {
+              remember_display = strdup (argv[1]);
+              if (!remember_display)
+                {
+                  fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
+                  exit (EXIT_FAILURE);
+                }
+            }
+          found = 1;
+          break;
+        }
+      else if (!strncmp (*argv, "--display=", 10))
+        {
+          if (!remember_display)
+            {
+              remember_display = strdup (*argv+10);
+              if (!remember_display)
+                {
+                  fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
+                  exit (EXIT_FAILURE);
+                }
+            }
+          found = 1;
+          break;
+        }
+    }
+
+  {
+    const char *s;
+    s = getenv ("DISPLAY");
+    if (s && *s)
+      found = 1;
+  }
+
+  return found;
 }
 
 
@@ -607,7 +819,6 @@ my_strusage( int level )
               {
                 snprintf (str, n, "Usage: %s [options] (-h for help)",
                           this_pgmname);
-                str[n-1] = 0;
               }
           }
         p = str;
@@ -695,6 +906,7 @@ pinentry_parse_opts (int argc, char *argv[])
                  "Grab keyboard only while window is focused"),
     ARGPARSE_s_u('W', "parent-wid", "Parent window ID (for positioning)"),
     ARGPARSE_s_s('c', "colors", "|STRING|Set custom colors for ncurses"),
+    ARGPARSE_s_s('a', "ttyalert", "|STRING|Set the alert mode (none, beep or flash)"),
     ARGPARSE_end()
   };
   ARGPARSE_ARGS pargs = { &argc, &argv, 0 };
@@ -720,9 +932,7 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.display = strdup (pargs.r.ret_str);
 	  if (!pinentry.display)
 	    {
-#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
-#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -730,19 +940,15 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.ttyname = strdup (pargs.r.ret_str);
 	  if (!pinentry.ttyname)
 	    {
-#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
-#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
 	case 'N':
-	  pinentry.ttytype = strdup (pargs.r.ret_str);
-	  if (!pinentry.ttytype)
+	  pinentry.ttytype_l = strdup (pargs.r.ret_str);
+	  if (!pinentry.ttytype_l)
 	    {
-#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
-#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -750,9 +956,7 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.lc_ctype = strdup (pargs.r.ret_str);
 	  if (!pinentry.lc_ctype)
 	    {
-#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
-#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -760,9 +964,7 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.lc_messages = strdup (pargs.r.ret_str);
 	  if (!pinentry.lc_messages)
 	    {
-#ifndef HAVE_W32CE_SYSTEM
 	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
-#endif
 	      exit (EXIT_FAILURE);
 	    }
 	  break;
@@ -779,6 +981,10 @@ pinentry_parse_opts (int argc, char *argv[])
             tmpstr = parse_color (tmpstr, &pinentry.color_bg, NULL);
             tmpstr = parse_color (tmpstr, &pinentry.color_so,
                                   &pinentry.color_so_bright);
+            tmpstr = parse_color (tmpstr, &pinentry.color_ok,
+                                  &pinentry.color_ok_bright);
+            tmpstr = parse_color (tmpstr, &pinentry.color_qualitybar,
+                                  &pinentry.color_qualitybar_bright);
           }
 	  break;
 
@@ -786,13 +992,38 @@ pinentry_parse_opts (int argc, char *argv[])
 	  pinentry.timeout = pargs.r.ret_int;
 	  break;
 
+	case 'a':
+	  pinentry.ttyalert = strdup (pargs.r.ret_str);
+	  if (!pinentry.ttyalert)
+	    {
+	      fprintf (stderr, "%s: %s\n", this_pgmname, strerror (errno));
+	      exit (EXIT_FAILURE);
+	    }
+	  break;
+
         default:
           pargs.err = ARGPARSE_PRINT_WARNING;
 	  break;
         }
     }
+
+  if (!pinentry.display && remember_display)
+    {
+      pinentry.display = remember_display;
+      remember_display = NULL;
+    }
 }
 
+
+/* Set the optional flag used with getinfo. */
+void
+pinentry_set_flavor_flag (const char *string)
+{
+  flavor_flag = string;
+}
+
+
+
 
 static gpg_error_t
 option_handler (assuan_context_t ctx, const char *key, const char *value)
@@ -805,7 +1036,7 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
     pinentry.grab = 1;
   else if (!strcmp (key, "debug-wait"))
     {
-#ifndef HAVE_W32CE_SYSTEM
+#ifndef HAVE_W32_SYSTEM
       fprintf (stderr, "%s: waiting for debugger - my pid is %u ...\n",
 	       this_pgmname, (unsigned int) getpid());
       sleep (*value?atoi (value):5);
@@ -830,10 +1061,18 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
     }
   else if (!strcmp (key, "ttytype"))
     {
-      if (pinentry.ttytype)
-	free (pinentry.ttytype);
-      pinentry.ttytype = strdup (value);
-      if (!pinentry.ttytype)
+      if (pinentry.ttytype_l)
+	free (pinentry.ttytype_l);
+      pinentry.ttytype_l = strdup (value);
+      if (!pinentry.ttytype_l)
+	return gpg_error_from_syserror ();
+    }
+  else if (!strcmp (key, "ttyalert"))
+    {
+      if (pinentry.ttyalert)
+	free (pinentry.ttyalert);
+      pinentry.ttyalert = strdup (value);
+      if (!pinentry.ttyalert)
 	return gpg_error_from_syserror ();
     }
   else if (!strcmp (key, "lc-ctype"))
@@ -852,6 +1091,46 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
       if (!pinentry.lc_messages)
 	return gpg_error_from_syserror ();
     }
+  else if (!strcmp (key, "owner"))
+    {
+      long along;
+      char *endp;
+
+      free (pinentry.owner_host);
+      pinentry.owner_host = NULL;
+      pinentry.owner_uid = -1;
+      pinentry.owner_pid = 0;
+
+      errno = 0;
+      along = strtol (value, &endp, 10);
+      if (along && !errno)
+        {
+          pinentry.owner_pid = (unsigned long)along;
+          if (*endp)
+            {
+              errno = 0;
+              if (*endp == '/') { /* we have a uid */
+                endp++;
+                along = strtol (endp, &endp, 10);
+                if (along >= 0 && !errno)
+                  pinentry.owner_uid = (int)along;
+              }
+              if (endp)
+                {
+                  while (*endp == ' ')
+                    endp++;
+                  if (*endp)
+                    {
+                      pinentry.owner_host = strdup (endp);
+                      for (endp=pinentry.owner_host;
+                           *endp && *endp != ' '; endp++)
+                        ;
+                      *endp = 0;
+                    }
+                }
+            }
+        }
+    }
   else if (!strcmp (key, "parent-wid"))
     {
       pinentry.parent_wid = atoi (value);
@@ -907,9 +1186,16 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
       if (!pinentry.default_tt_hide)
 	return gpg_error_from_syserror ();
     }
+  else if (!strcmp (key, "default-capshint"))
+    {
+      pinentry.default_capshint = strdup (value);
+      if (!pinentry.default_capshint)
+	return gpg_error_from_syserror ();
+    }
   else if (!strcmp (key, "allow-external-password-cache") && !*value)
     {
-      pinentry.allow_external_password_cache = 1;
+      char *desktop = getenv ("XDG_SESSION_DESKTOP");
+      pinentry.allow_external_password_cache = (!desktop || strcmp (desktop, "KDE"));
       pinentry.tried_password_cache = 0;
     }
   else if (!strcmp (key, "allow-emacs-prompt") && !*value)
@@ -926,6 +1212,48 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
       if (!pinentry.invisible_char)
 	return gpg_error_from_syserror ();
     }
+  else if (!strcmp (key, "formatted-passphrase") && !*value)
+    {
+      pinentry.formatted_passphrase = 1;
+    }
+  else if (!strcmp (key, "formatted-passphrase-hint"))
+    {
+      if (pinentry.formatted_passphrase_hint)
+        free (pinentry.formatted_passphrase_hint);
+      pinentry.formatted_passphrase_hint = strdup (value);
+      if (!pinentry.formatted_passphrase_hint)
+	return gpg_error_from_syserror ();
+      do_unescape_inplace(pinentry.formatted_passphrase_hint);
+    }
+  else if (!strcmp (key, "constraints-enforce") && !*value)
+    pinentry.constraints_enforce = 1;
+  else if (!strcmp (key, "constraints-hint-short"))
+    {
+      if (pinentry.constraints_hint_short)
+        free (pinentry.constraints_hint_short);
+      pinentry.constraints_hint_short = strdup (value);
+      if (!pinentry.constraints_hint_short)
+	return gpg_error_from_syserror ();
+      do_unescape_inplace(pinentry.constraints_hint_short);
+    }
+  else if (!strcmp (key, "constraints-hint-long"))
+    {
+      if (pinentry.constraints_hint_long)
+        free (pinentry.constraints_hint_long);
+      pinentry.constraints_hint_long = strdup (value);
+      if (!pinentry.constraints_hint_long)
+	return gpg_error_from_syserror ();
+      do_unescape_inplace(pinentry.constraints_hint_long);
+    }
+  else if (!strcmp (key, "constraints-error-title"))
+    {
+      if (pinentry.constraints_error_title)
+        free (pinentry.constraints_error_title);
+      pinentry.constraints_error_title = strdup (value);
+      if (!pinentry.constraints_error_title)
+	return gpg_error_from_syserror ();
+      do_unescape_inplace(pinentry.constraints_error_title);
+    }
   else
     return gpg_error (GPG_ERR_UNKNOWN_OPTION);
   return 0;
@@ -969,7 +1297,6 @@ write_status_error (assuan_context_t ctx, pinentry_t pe)
             pe->specific_err_loc? pe->specific_err_loc : "?",
             pe->specific_err,
             pe->specific_err_info? pe->specific_err_info : "");
-  buf[sizeof buf -1] = 0;
   assuan_write_status (ctx, "ERROR", buf);
 }
 
@@ -1050,6 +1377,23 @@ cmd_setrepeat (assuan_context_t ctx, char *line)
   return 0;
 }
 
+static gpg_error_t
+cmd_setrepeatok (assuan_context_t ctx, char *line)
+{
+  char *p;
+
+  (void)ctx;
+
+  p = malloc (strlen (line) + 1);
+  if (!p)
+    return gpg_error_from_syserror ();
+
+  strcpy_escaped (p, line);
+  free (pinentry.repeat_ok_string);
+  pinentry.repeat_ok_string = p;
+  return 0;
+}
+
 
 static gpg_error_t
 cmd_setrepeaterror (assuan_context_t ctx, char *line)
@@ -1219,6 +1563,53 @@ cmd_setqualitybar_tt (assuan_context_t ctx, char *line)
   return 0;
 }
 
+/* Set the tooltip to be used for a generate action.  */
+static gpg_error_t
+cmd_setgenpin_tt (assuan_context_t ctx, char *line)
+{
+  char *newval;
+
+  (void)ctx;
+
+  if (*line)
+    {
+      newval = malloc (strlen (line) + 1);
+      if (!newval)
+        return gpg_error_from_syserror ();
+
+      strcpy_escaped (newval, line);
+    }
+  else
+    newval = NULL;
+  if (pinentry.genpin_tt)
+    free (pinentry.genpin_tt);
+  pinentry.genpin_tt = newval;
+  return 0;
+}
+
+/* Set the label to be used for a generate action.  */
+static gpg_error_t
+cmd_setgenpin_label (assuan_context_t ctx, char *line)
+{
+  char *newval;
+
+  (void)ctx;
+
+  if (*line)
+    {
+      newval = malloc (strlen (line) + 1);
+      if (!newval)
+        return gpg_error_from_syserror ();
+
+      strcpy_escaped (newval, line);
+    }
+  else
+    newval = NULL;
+  if (pinentry.genpin_label)
+    free (pinentry.genpin_label);
+  pinentry.genpin_label = newval;
+  return 0;
+}
 
 static gpg_error_t
 cmd_getpin (assuan_context_t ctx, char *line)
@@ -1233,6 +1624,8 @@ cmd_getpin (assuan_context_t ctx, char *line)
   if (!pinentry.pin)
     return gpg_error (GPG_ERR_ENOMEM);
 
+  pinentry.confirm = 0;
+
   /* Try reading from the password cache.  */
   if (/* If repeat passphrase is set, then we don't want to read from
 	 the cache.  */
@@ -1247,10 +1640,14 @@ cmd_getpin (assuan_context_t ctx, char *line)
       && ! pinentry.error)
     {
       char *password;
+      int give_up_on_password_store = 0;
 
       pinentry.tried_password_cache = 1;
 
-      password = password_cache_lookup (pinentry.keyinfo);
+      password = password_cache_lookup (pinentry.keyinfo, &give_up_on_password_store);
+      if (give_up_on_password_store)
+	pinentry.allow_external_password_cache = 0;
+
       if (password)
 	/* There is a cached password.  Try it.  */
 	{
@@ -1321,6 +1718,10 @@ cmd_getpin (assuan_context_t ctx, char *line)
       if (pinentry.specific_err)
         {
           write_status_error (ctx, &pinentry);
+
+          if (gpg_err_code (pinentry.specific_err) == GPG_ERR_FULLY_CANCELED)
+            assuan_set_flag (ctx, ASSUAN_FORCE_CLOSE, 1);
+
           return pinentry.specific_err;
         }
       return (pinentry.locale_err
@@ -1333,9 +1734,11 @@ cmd_getpin (assuan_context_t ctx, char *line)
     {
       if (pinentry.repeat_okay)
         assuan_write_status (ctx, "PIN_REPEATED", "");
+      assuan_begin_confidential (ctx);
       result = assuan_send_data (ctx, pinentry.pin, strlen(pinentry.pin));
       if (!result)
 	result = assuan_send_data (ctx, NULL, 0);
+      assuan_end_confidential (ctx);
 
       if (/* GPG Agent says it's okay.  */
 	  pinentry.allow_external_password_cache && pinentry.keyinfo
@@ -1374,6 +1777,7 @@ cmd_confirm (assuan_context_t ctx, char *line)
   free (pinentry.specific_err_info);
   pinentry.specific_err_info = NULL;
   pinentry.canceled = 0;
+  pinentry.confirm = 1;
   pinentry_setbuffer_clear (&pinentry);
   result = (*pinentry_cmd_handler) (&pinentry);
   if (pinentry.error)
@@ -1385,12 +1789,16 @@ cmd_confirm (assuan_context_t ctx, char *line)
   if (pinentry.close_button)
     assuan_write_status (ctx, "BUTTON_INFO", "close");
 
-  if (result)
-    return 0;
+  if (result > 0)
+    return 0; /* OK */
 
   if (pinentry.specific_err)
     {
       write_status_error (ctx, &pinentry);
+
+      if (gpg_err_code (pinentry.specific_err) == GPG_ERR_FULLY_CANCELED)
+        assuan_set_flag (ctx, ASSUAN_FORCE_CLOSE, 1);
+
       return pinentry.specific_err;
     }
 
@@ -1398,7 +1806,7 @@ cmd_confirm (assuan_context_t ctx, char *line)
     return gpg_error (GPG_ERR_LOCALE_PROBLEM);
 
   if (pinentry.one_button)
-    return 0;
+    return 0; /* OK */
 
   if (pinentry.canceled)
     return gpg_error (GPG_ERR_CANCELED);
@@ -1414,6 +1822,33 @@ cmd_message (assuan_context_t ctx, char *line)
   return cmd_confirm (ctx, "--one-button");
 }
 
+
+/* Return a staically allocated string with information on the mode,
+ * uid, and gid of DEVICE.  On error "?" is returned if DEVICE is
+ * NULL, "-" is returned.  */
+static const char *
+device_stat_string (const char *device)
+{
+#ifdef HAVE_STAT
+  static char buf[40];
+  struct stat st;
+
+  if (!device || !*device)
+    return "-";
+
+  if (stat (device, &st))
+    return "?";  /* Error */
+  snprintf (buf, sizeof buf, "%lo/%lu/%lu",
+            (unsigned long)st.st_mode,
+            (unsigned long)st.st_uid,
+            (unsigned long)st.st_gid);
+  return buf;
+#else
+  return "-";
+#endif
+}
+
+
 /* GETINFO <what>
 
    Multipurpose function to return a variety of information.
@@ -1422,13 +1857,14 @@ cmd_message (assuan_context_t ctx, char *line)
      version     - Return the version of the program.
      pid         - Return the process id of the server.
      flavor      - Return information about the used pinentry flavor
+     ttyinfo     - Return DISPLAY, ttyinfo and an emacs pinentry status
  */
 static gpg_error_t
 cmd_getinfo (assuan_context_t ctx, char *line)
 {
   int rc;
   const char *s;
-  char buffer[100];
+  char buffer[150];
 
   if (!strcmp (line, "version"))
     {
@@ -1439,33 +1875,44 @@ cmd_getinfo (assuan_context_t ctx, char *line)
     {
 
       snprintf (buffer, sizeof buffer, "%lu", (unsigned long)getpid ());
-      buffer[sizeof buffer -1] = 0;
       rc = assuan_send_data (ctx, buffer, strlen (buffer));
     }
   else if (!strcmp (line, "flavor"))
     {
-      const char *flags;
-
       if (!strncmp (this_pgmname, "pinentry-", 9) && this_pgmname[9])
         s = this_pgmname + 9;
       else
         s = this_pgmname;
 
-      if (0)
-        ;
+      snprintf (buffer, sizeof buffer, "%s%s%s",
+                s,
+                flavor_flag? ":":"",
+                flavor_flag? flavor_flag : "");
+      rc = assuan_send_data (ctx, buffer, strlen (buffer));
+      /* if (!rc) */
+      /*   rc = assuan_write_status (ctx, "FEATURES", "tabbing foo bar"); */
+    }
+  else if (!strcmp (line, "ttyinfo"))
+    {
+      char emacs_status[10];
 #ifdef INSIDE_EMACS
-      else if (pinentry_cmd_handler == emacs_cmd_handler)
-        flags = ":emacs";
+      snprintf (emacs_status, sizeof emacs_status,
+                "%d", pinentry_emacs_status ());
+#else
+      strcpy (emacs_status, "-");
 #endif
-#ifdef FALLBACK_CURSES
-      else if (pinentry_cmd_handler == curses_cmd_handler)
-        flags = ":curses";
+      snprintf (buffer, sizeof buffer, "%s %s %s %s %lu/%lu %s",
+                pinentry.ttyname? pinentry.ttyname : "-",
+                pinentry.ttytype_l? pinentry.ttytype_l : "-",
+                pinentry.display? pinentry.display : "-",
+                device_stat_string (pinentry.ttyname),
+#ifdef HAVE_DOSISH_SYSTEM
+                0l, 0l,
+#else
+                (unsigned long)geteuid (), (unsigned long)getegid (),
 #endif
-      else
-        flags = "";
-
-      snprintf (buffer, sizeof buffer, "%s%s", s, flags);
-      buffer[sizeof buffer -1] = 0;
+                emacs_status
+                );
       rc = assuan_send_data (ctx, buffer, strlen (buffer));
     }
   else
@@ -1515,6 +1962,7 @@ register_commands (assuan_context_t ctx)
       { "SETKEYINFO", cmd_setkeyinfo },
       { "SETREPEAT",  cmd_setrepeat },
       { "SETREPEATERROR", cmd_setrepeaterror },
+      { "SETREPEATOK", cmd_setrepeatok},
       { "SETERROR",   cmd_seterror },
       { "SETOK",      cmd_setok },
       { "SETNOTOK",   cmd_setnotok },
@@ -1524,6 +1972,8 @@ register_commands (assuan_context_t ctx)
       { "MESSAGE",    cmd_message },
       { "SETQUALITYBAR", cmd_setqualitybar },
       { "SETQUALITYBAR_TT", cmd_setqualitybar_tt },
+      { "SETGENPIN",    cmd_setgenpin_label },
+      { "SETGENPIN_TT", cmd_setgenpin_tt },
       { "GETINFO",    cmd_getinfo },
       { "SETTITLE",   cmd_settitle },
       { "SETTIMEOUT", cmd_settimeout },
